/**************************************************************************
*  This file is part of the TAL project (Tiny Abstraction Layer)
*
*  Copyright (c) 2013-2016 by Michael Fischer (www.emb4fun.de).
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
***************************************************************************
*  History:
*
*  20.12.2013  mifi  First Version
**************************************************************************/
#define __TALCPU_C__

/*=======================================================================*/
/*  Include                                                              */
/*=======================================================================*/
#include "tal.h"

/*=======================================================================*/
/*  All Structures and Common Constants                                  */
/*=======================================================================*/
typedef void (*IRQ_FUNC)(void);

#define SYSIRQ_PRIO  7

#define ARM7_IRQ_ENTRY()                                                                  \
{                                                                                         \
   __asm__ volatile ( "sub    lr, lr,#4\n"            /* Adjust LR */                     \
                      "stmfd  sp!,{r0-r12,lr}\n"      /* Save registers on IRQ stack. */  \
                      "mrs    r1, spsr\n"             /* Save SPSR */                     \
                      "stmfd  sp!,{r1}" );                                                \
}

#define ARM7_IRQ_EXIT()                                                                   \
{                                                                                         \
   __asm__ volatile ( "ldmfd  sp!, {r1}\n"            /* Restore SPSR */                  \
                      "msr    spsr_c, r1\n"                                               \
                      "ldr    r0, =0xFFFFF000\n"      /* End of interrupt. */             \
                      "str    r0, [r0, #0x130]\n"                                         \
                      "ldmfd  sp!, {r0-r12, pc}^" );  /* Restore registers and return. */ \
}



/*=======================================================================*/
/*  Definition of all local Data                                         */
/*=======================================================================*/

static uint32_t dHiResPeriod = 0;

/*=======================================================================*/
/*  Definition of all local Procedures                                   */
/*=======================================================================*/

/*************************************************************************/
/*  SpuriousHandler                                                      */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void SpuriousHandler (void)
{
} /* SpuriousHandler */

/*************************************************************************/
/*  SYSIrqHandler                                                        */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void SYSIrqHandler (void)
{
   /* Check Periodic Interval Timer Controller */
   if (AT91C_BASE_PITC->PITC_PISR & AT91C_PITC_PITS)
   {
      (void)AT91C_BASE_PITC->PITC_PIVR;
      OSTimerCallback();
   }
} /* SYSIrqHandler */

/*=======================================================================*/
/*  All code exported                                                    */
/*=======================================================================*/

/*************************************************************************/
/*  InitSysTick                                                          */
/*                                                                       */
/*  Init the SysTick timer.                                              */
/*                                                                       */                                                                       
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void InitSysTick (void)
{
   static uint8_t InitDone = 0;
   
   if (0 == InitDone)
   {
      InitDone = 1;
      
      /* Disable all interrupts first */
      AT91C_BASE_AIC->AIC_IDCR = 0xFFFFFFFF;
   
      /* Set end of interrupt command */
      AT91C_BASE_AIC->AIC_EOICR = 0;

      /* Set spurious interrupt handler */ 
      AT91C_BASE_AIC->AIC_SPU = (unsigned long)SpuriousHandler;
   
      /*
       * The CPU clock is set to 48MHz by the startup code
       * from CrossWorks for ARM. Here we must only init the
       * timer. Setup the Periodic Interval Timer with a period
       * of OS_TICKS_PER_SECOND.
       */
      dHiResPeriod = ((tal_CPUGetFrequencyCPU() / 16) / OS_TICKS_PER_SECOND) - 1;

      /* 
       * PIT and PIT Interrupt must be
       * enabled with tal_CPUSysTickStart later.
       */
       
      /*
       * Install the SystemIRQHandler, which will be used
       * for the timer interrupt.
       *
       * Setup System Interrupt Mode and Vector.
       * PIT's int is System Interrupt, Peripheral ID for PIT is 1
       */
      AT91C_BASE_AIC->AIC_SMR[1] = (unsigned long)SYSIRQ_PRIO;
      AT91C_BASE_AIC->AIC_SVR[1] = (unsigned long)SYSIrqHandler;
   }      
} /* InitSysTick */

/*************************************************************************/
/*  tal_CPUInit                                                          */
/*                                                                       */
/*  "Initialize" the CPU.                                                */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void tal_CPUInit (void)
{
   /* Init system ticker */
   InitSysTick();
   
   /* Enable System Peripheral INT */
   AT91C_BASE_AIC->AIC_IECR = (1<<1);   
} /* tal_CPUInit */

/*************************************************************************/
/*  tal_CPUSysTickStart                                                  */
/*                                                                       */
/*  Start the SysTick.                                                   */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void tal_CPUSysTickStart (void)
{
   AT91C_BASE_PITC->PITC_PIMR = (1 << 24) |    /* PIT Enable */
                                (1 << 25) |    /* PIT Interrupt Enable */
                                dHiResPeriod;
} /* tal_CPUSysTickStart */

/*************************************************************************/
/*  tal_CPUIrqEnable                                                     */
/*                                                                       */
/*  Enable the given IRQ.                                                */
/*                                                                       */
/*  In    : IRQ                                                          */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void tal_CPUIrqEnable (int IRQ)
{
   if (IRQ <= AT91C_ID_31_Reserved)
   {
      AT91C_BASE_AIC->AIC_IECR = (1<<IRQ);
   }
} /* tal_CPUIrqEnable */

/*************************************************************************/
/*  tal_CPUIrqDisable                                                    */
/*                                                                       */
/*  Disable the given IRQ.                                               */
/*                                                                       */
/*  In    : IRQ                                                          */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void tal_CPUIrqDisable (int IRQ)
{
   if (IRQ <= AT91C_ID_31_Reserved)
   {
      AT91C_BASE_AIC->AIC_IDCR = (1<<IRQ);
   }
} /* tal_CPUIrqDisable */

/*************************************************************************/
/*  tal_CPUIrqSetPriority                                                */
/*                                                                       */
/*  Set priority of the given IRQ.                                       */
/*                                                                       */
/*  In    : IRQ, Priority                                                */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void tal_CPUIrqSetPriority (int IRQ, int Priority)
{
   if (IRQ <= AT91C_ID_31_Reserved)
   {
      AT91C_BASE_AIC->AIC_SMR[IRQ] = Priority;   
   }
} /* tal_CPUIrqSetPriority */

/*************************************************************************/
/*  tal_CPUIrqSetFunction                                                */
/*                                                                       */
/*  Set IRQ function of the given IRQ.                                   */
/*                                                                       */
/*  In    : IRQ, Function                                                */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void tal_CPUIrqSetFunction (int IRQ, tal_irq_fnt Function)
{
   if (IRQ <= AT91C_ID_31_Reserved)
   {
      AT91C_BASE_AIC->AIC_SVR[IRQ] = (uint32_t)Function;   
   }
} /* tal_CPUIrqSetFunction */

/*************************************************************************/
/*  tal_CPUStatGetHiResPeriod                                            */
/*                                                                       */
/*  Return the HiResPeriod value.                                        */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: HiResPeriod                                                  */
/*************************************************************************/
uint32_t tal_CPUStatGetHiResPeriod (void)
{
   return(dHiResPeriod);
} /* tal_CPUStatGetHiResPeriod */

/*************************************************************************/
/*  tal_CPUStatGetHiResCnt                                               */
/*                                                                       */
/*  Return the HiRes counter.                                            */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: HiRes counter                                                */
/*************************************************************************/
uint32_t tal_CPUStatGetHiResCnt (void)
{
   uint32_t dValue;
   
   /* Get milliseconds */
   dValue  = (OSTimeGet() << 16);

   dValue |= (uint16_t)AT91C_BASE_SYS->PITC_PIIR;
   
   return(dValue);
} /* tal_CPUStatGetHiResCnt */

/*************************************************************************/
/*  tal_CPUGetFrequencyCPU                                               */
/*                                                                       */
/*  Return the clock frequency of the CPU in MHz.                        */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: Frequency                                                    */
/*************************************************************************/
uint32_t tal_CPUGetFrequencyCPU (void)
{
   uint32_t dFrequency = 48054857;
   
   return(dFrequency);
} /* tal_CPUGetFrequencyCPU */

/*************************************************************************/
/*  irq_handler                                               (CTSonly)  */
/*                                                                       */
/*  This is the ARM7 irq_handler.                                        */
/*                                                                       */                                                                       
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void irq_handler (void) __attribute__((__naked__));
void irq_handler (void)
{
   /* Interrupt prologue */
   __asm__ volatile ( "sub    lr, lr,#4\n"         /* Adjust LR */                    
                      "stmfd  sp!,{r0-r12,lr}\n"   /* Save registers on IRQ stack. */ 
                      "mrs    r1, spsr\n"          /* Save SPSR */                    
                      "stmfd  sp!,{r1}" );                                               
   /********************************************/                      
                         
   IRQ_FUNC IrqFunc;
   
   /* Get the interrupt source */   
   IrqFunc = (IRQ_FUNC)AT91C_BASE_AIC->AIC_IVR;   
   if (IrqFunc != (IRQ_FUNC)0)
   {
      IrqFunc();
   }

   /* Interrupt epilogue */
   __asm__ volatile ( "ldmfd  sp!, {r1}\n"            /* Restore SPSR */
                      "msr    spsr_c, r1\n"
                      "ldr    r0, =0xFFFFF000\n"      /* End of interrupt. */
                      "str    r0, [r0, #0x130]\n"
                      "ldmfd  sp!, {r0-r12, pc}^" );  /* Restore registers and return. */
} /* irq_handler */

/*** EOF ***/
