/**************************************************************************
*  This file is part of the TAL project (Tiny Abstraction Layer)
*
*  Copyright (c) 2013-2016 by Michael Fischer (www.emb4fun.de).
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
***************************************************************************
*  History:
*
*  20.12.2013  mifi  First Version
**************************************************************************/
#define __TALCPU_COM_C__

/*=======================================================================*/
/*  Include                                                              */
/*=======================================================================*/
#include "tal.h"

/*=======================================================================*/
/*  All Structures and Common Constants                                  */
/*=======================================================================*/

/*=======================================================================*/
/*  Definition of all local Data                                         */
/*=======================================================================*/

static TAL_COM_DCB *DCBArray[TAL_COM_PORT_MAX];

/*=======================================================================*/
/*  Definition of all local Procedures                                   */
/*=======================================================================*/

/*************************************************************************/
/*  IRQHandler                                                           */
/*                                                                       */
/*  This is the generic IRQ handler.                                     */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void IRQHandler (TAL_COM_DCB *pDCB)
{
   TAL_RESULT    Error;
   AT91S_USART *pUSARTx;
   uint32_t     dStatus;
   uint8_t      bData;
   
   /* Get USART pointer */
   pUSARTx = (AT91S_USART *)pDCB->HW.dBaseAddress;

   /* Get IRQ status */   
   dStatus = pUSARTx->US_CSR;
   
   /* 
    * Check for RX interrupt
    * "Read data register not empty" and "Overrun error"  
    */
   if (dStatus & (AT91C_US_RXRDY | AT91C_US_OVRE))
   {
      /* Read data */
      bData = (uint8_t)pUSARTx->US_RHR;
      
      /* If we have no overflow... */
      if (TAL_FALSE == pDCB->bRxOverflow)
      {
         /* ... put it into the ring buffer */
         Error = tal_MISCRingAdd(&pDCB->RxRing, &bData);
         if (TAL_OK == Error)
         {
            /* Signal counting semaphore */
            OSSemaSignalFromInt(&pDCB->RxRdySema);
         }
         else
         {
            /* Ups, overflow */
            pDCB->bRxOverflow = TAL_OK;
         }
         
         /* Signal data available */
         OSSemaSignalFromInt(&pDCB->RxRdySema);
      }  
   } /* end "RX interrupt" */
   
   /* 
    * Check for TX interrupt, but only if it is enabled 
    */
   if ( (pUSARTx->US_IMR & AT91C_US_TXRDY) && (dStatus & AT91C_US_TXRDY) )
   {
      /* Read Data from the ring buffer */
      Error = tal_MISCRingGet(&pDCB->TxRing, &bData);
      if (Error != TAL_OK)
      {
         /* Ups, no data available, disable interrupt */
         pUSARTx->US_IDR = AT91C_US_TXRDY;
      }
      else
      {
         /* Send data */
         pUSARTx->US_THR = bData;
      }   
   } /* end "TX interrupt */
   
} /* IRQHandler */

/*************************************************************************/
/*  USARTx_IRQHandler                                                    */
/*                                                                       */
/*  This is the ARM7 USARTx IRQ handler.                                 */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void USART1_IRQHandler (void)
{
   TAL_CPU_IRQ_ENTER();
   IRQHandler(DCBArray[TAL_COM_PORT_1]);
   TAL_CPU_IRQ_EXIT();
} /* USART1_IRQHandler */

/*=======================================================================*/
/*  All code exported                                                    */
/*=======================================================================*/

/*************************************************************************/
/*  cpu_COMInit                                                          */
/*                                                                       */
/*  Prepare the hardware for use by the Open function later. Set the HW  */
/*  information depending of ePort and "enable" the COM port.            */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_COMInit (TAL_COM_DCB *pDCB)
{
   TAL_RESULT    Error = TAL_ERR_COM_PORT_RANGE;
   AT91S_USART *pUSARTx;
   TAL_COM_HW  *pHW = &pDCB->HW;
   
   switch (pDCB->ePort)
   {
      case TAL_COM_PORT_1:
      {
         Error = tal_BoardEnableCOM1();
         if (TAL_OK == Error)
         {
            DCBArray[TAL_COM_PORT_1] = pDCB;
            
            pHW->dBaseAddress = (uint32_t)AT91C_BASE_US0;
            pHW->IrqNumber    = AT91C_ID_US0;
            pHW->bIrqPriority = USART0_PRIO;
            pHW->bOver8       = USART0_OVER_8;
            
            /* Set interrupt priority, "High level Sensitive" */
            AT91C_BASE_AIC->AIC_SMR[pHW->IrqNumber] = ((0 << 5) | pHW->bIrqPriority); 
            
            /* Set the interrupt vector */
            AT91C_BASE_AIC->AIC_SVR[pHW->IrqNumber] = (uint32_t)USART1_IRQHandler;
            
            /* Get USART pointer */
            pUSARTx = (AT91S_USART *)pDCB->HW.dBaseAddress;
            
            /* Reset transmitter and receiver */
            pUSARTx->US_CR = AT91C_US_RSTTX | AT91C_US_RSTRX;
         }
         break;
      } /* TAL_COM_PORT_1 */
      
      default:
      {
         /* Do nothing */
         break;
      }
   } /* end switch (pDCB->ePort) */
   
   return(Error);
} /* cpu_COMInit */

/*************************************************************************/
/*  cpu_COMOpen                                                          */
/*                                                                       */
/*  Open the COM port.                                                   */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_COMOpen (TAL_COM_DCB *pDCB)
{
   TAL_RESULT    Error = TAL_ERROR;
   AT91S_USART *pUSARTx;
   TAL_COM_HW  *pHW   = &pDCB->HW;
   uint32_t     dMode = 0;
   uint32_t     dBRR  = 0;
   uint8_t      bDummy;

   /* Get USART pointer */
   pUSARTx = (AT91S_USART *)pDCB->HW.dBaseAddress;

   /* Check word length */
   switch (pDCB->Settings.eLength)
   {
      case TAL_COM_LENGTH_8:
      {
         dMode |= AT91C_US_CHRL_8_BITS;
         break;
      }
      
#if 0 /* Not supported by the ringbuffer  for the moment */      
      case TAL_COM_LENGTH_9:
      {
         dMode |= AT91C_US_MODE9;
         break;
      }
#endif      
      
      default:
      {
         Error = TAL_ERR_COM_LENGTH;
         goto COMOpenEnd;
         break;   /*lint !e527*/
      }
   } /* switch (pDCB->Settings.eLength) */
   
   /* Check parity settings */
   switch (pDCB->Settings.eParity)
   {
      case TAL_COM_PARITY_NONE:
      {
         dMode |= (4<<9);
         break;
      } 
      
      case TAL_COM_PARITY_EVEN:
      {
         dMode |= (0<<9);
         break;
      }
      
      case TAL_COM_PARITY_ODD:
      {
         dMode |= (1<<9);
         break;
      }
      
      default:
      {
         Error = TAL_ERR_COM_PARITY;
         goto COMOpenEnd;
         break;   /*lint !e527*/
      }
   } /* switch (pDCB->Settings.eParity) */
   
   /* Check stop bit settings */
   switch (pDCB->Settings.eStop)
   {
      case TAL_COM_STOP_1_0:
      {
         dMode |= AT91C_US_NBSTOP_1_BIT;
         break;
      }
      
      case TAL_COM_STOP_1_5:
      {
         dMode |= AT91C_US_NBSTOP_15_BIT;
         break;
      }
      
      case TAL_COM_STOP_2_0:
      {
         dMode |= AT91C_US_NBSTOP_2_BIT;
         break;
      }
      
      default:
      {
         Error = TAL_ERR_COM_STOP;
         goto COMOpenEnd;
         break;   /*lint !e527*/
      }
   } /* switch (pDCB->Settings.eStop) */
   
   /* Check baud rate */
   if (pDCB->Settings.dBaudrate != 0)
   {
      if (TAL_TRUE == pDCB->HW.bOver8)
      {
         dMode |= AT91C_US_OVER;   
         dBRR   = tal_CPUGetFrequencyCPU() / (pDCB->Settings.dBaudrate * 8);
      }
      else
      {
         dBRR   = tal_CPUGetFrequencyCPU() / (pDCB->Settings.dBaudrate * 16);
      }   
   }
   else
   {
      Error = TAL_ERR_COM_BAUDRATE;
      goto COMOpenEnd;
   }
   
   /* Enable transmitter and receiver */
   pUSARTx->US_CR = AT91C_US_TXEN | AT91C_US_RXEN;
   
   /* Configure USART */
   pUSARTx->US_MR   = dMode;
   pUSARTx->US_BRGR = dBRR;
   
   /* Dummy read to clear the DR register */
   bDummy = (uint8_t)pUSARTx->US_RHR;
   (void)bDummy;
   
   /* Enable RX interrrupt */
   pUSARTx->US_IER = AT91C_US_RXRDY;
   
   /* Enables corresponding USARTx interrupt. */
   AT91C_BASE_AIC->AIC_IECR = (1<<pHW->IrqNumber);
   
   Error = TAL_OK;

COMOpenEnd:   
   return(Error);
} /* cpu_COMOpen */

/*************************************************************************/
/*  cpu_COMClose                                                         */
/*                                                                       */
/*  Close the COM port.                                                  */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_COMClose (TAL_COM_DCB *pDCB)
{
   TAL_RESULT    Error = TAL_OK;
   AT91S_USART *pUSARTx;
   
   /* Get USART pointer */
   pUSARTx = (AT91S_USART *)pDCB->HW.dBaseAddress;

   /* Enables corresponding USARTx interrupt. */
   AT91C_BASE_AIC->AIC_IDCR = (1<<pDCB->HW.IrqNumber);

   /* Reset transmitter and receiver */
   pUSARTx->US_CR = AT91C_US_RSTTX | AT91C_US_RSTRX;
   
   return(Error);
} /* cpu_COMOpen */

/*************************************************************************/
/*  cpu_COMStartTx                                                       */
/*                                                                       */
/*  Send the data from the ring buffer if the TX interrupt is disabled.  */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_COMStartTx (TAL_COM_DCB *pDCB)
{
   TAL_RESULT    Error = TAL_OK;
   AT91S_USART *pUSARTx;
   uint8_t       bData;
   
   /* Get USART pointer */
   pUSARTx = (AT91S_USART *)pDCB->HW.dBaseAddress;

   TAL_CPU_DISABLE_ALL_INTS();
   if (pUSARTx->US_IMR & AT91C_US_TXRDY)
   {
      /* TX interrupt is enabled, do nothing */
   }
   else
   {
      /* Get data from the ring buffer */
      Error = tal_MISCRingGet(&pDCB->TxRing, &bData);
      if (TAL_OK == Error)
      {
         /* Send data */
         pUSARTx->US_THR = bData;
      
         /* Enable TX interrupt */
         pUSARTx->US_IER = AT91C_US_TXRDY;
      }
   }
   TAL_CPU_ENABLE_ALL_INTS();
   
   return(Error);
} /* cpu_COMStartTx */

/*************************************************************************/
/*  cpu_COMTxIsRunning                                                   */
/*                                                                       */
/*  Check if TX is still running.                                        */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / TAL_ERROR                                           */
/*************************************************************************/
TAL_RESULT cpu_COMTxIsRunning (TAL_COM_DCB *pDCB)
{
   TAL_RESULT    Error = TAL_OK;
   AT91S_USART *pUSARTx;
   
   /* Get USART pointer */
   pUSARTx = (AT91S_USART *)pDCB->HW.dBaseAddress;

   TAL_CPU_DISABLE_ALL_INTS();
   if (pUSARTx->US_IMR & AT91C_US_TXRDY)
   {
      /* TX is still running */
      Error = TAL_OK;
   }
   else
   {
      /* TX is not running */
      Error = TAL_OK;
   }
   TAL_CPU_ENABLE_ALL_INTS();
   
   return(Error);
} /* cpu_COMTxIsRunning */

/*** EOF ***/
